#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>

#include "udict.h"

typedef struct dmem_t {
    void *addr;
    size_t size;
    const char *file;
    int line;
} DMEM_T;

typedef enum {
    NOT_INITIALIZED,
    INITIALIZED,
    FAILED_TO_INITIALIZE
} INIT_STAT;

static INIT_STAT initialized = NOT_INITIALIZED;
static FILE *reportFp = NULL;
static udict_t *dict = NULL;
static int dumpMemory = 0;
static int dumpLeaks = 1;
static int noFinalize = 0;


static DMEM_T *dmem_t_new(const char *fn,int line,void *ptr,size_t size);
static void dmem_t_free(DMEM_T *t);
static void dmem_initialize(void);
static void dmem_finalize(void);
static void dmem_report(FILE *fp);


static DMEM_T *dmem_t_new(const char *fn,int line,void *ptr,size_t size)
{
    DMEM_T *ret = (DMEM_T *)malloc(sizeof(DMEM_T)*1);
    if ( ret == NULL ) return NULL;
    ret->addr = ptr;
    ret->file = fn;
    ret->line = line;
    ret->size = size;
    return ret;
}

static void dmem_t_free(DMEM_T *t)
{
    if ( t == NULL ) return;
    t->addr = NULL;
    t->file = NULL;
    free(t);
    // we can reuse the element :-(
}

static int comparef(const void *key1, const void *key2)
{
    return key1 - key2;
}

static void dmem_initialize(void)
{
    char *logFile = NULL;
    char *opts = NULL;

    if ( initialized != NOT_INITIALIZED ) return;

    initialized = INITIALIZED;

    logFile = getenv("DMEM_FILE");
    opts = getenv("DMEM_OPTION");

    if ( logFile != NULL && strlen(logFile) > 0 ) {
        reportFp = fopen(logFile,"wt");
        if ( reportFp == NULL ) {
            fprintf(stderr,"Unable to open file: %s\n",logFile);
        }
    }
    if ( reportFp == NULL ) reportFp = stderr;

    if ( opts != NULL && strlen(opts) > 0 ) {
        char *p;
        char *tmp;
        p = strtok_r(opts,":",&tmp);
        do {
            if ( !strcmp(p,"DUMP_MEMORY") ) {
                dumpMemory = 1;
            } else if ( !strcmp(p,"NO_LEAKS") ) {
                dumpLeaks = 0;
            } else if ( !strcmp(p,"NO_FINALIZE") ) {
                noFinalize = 1;
            }
        } while ( (p = strtok_r(NULL,":",&tmp)) != NULL );
    }

    /* initialize report file pointer */
    dict = udict_create(UDICT_AVL, UDICT_COUNT_T_MAX, comparef);
    if ( dict == NULL ) {
        fprintf(stderr,"failed to create data structure for tracking memory allocation\n");
        initialized = FAILED_TO_INITIALIZE;
    } else if ( atexit(dmem_finalize) ) {
        fprintf(stderr,"failed to install exit handler\n");
        initialized = FAILED_TO_INITIALIZE;
    }
}

static void dmem_finalize(void)
{
    DMEM_T *d;
    udict_node_t *n;

    if ( initialized != INITIALIZED ) return;

    /* generate report */
    dmem_report(reportFp);

    /* destroy everything */
    if ( reportFp != NULL && reportFp != stderr )
        fclose(reportFp);

    if ( !noFinalize ) {
        while ( !udict_isempty(dict) ) {
            n = udict_first(dict);
            d = (DMEM_T *)udict_node_getdata(n);
            udict_delete_free(dict,n);
            dmem_t_free(d);
        }
        udict_destroy(dict);
    }
}

static void dump_memory(FILE *fp,void *ptr,size_t s)
{
    register int i = 0;
    register unsigned char *p = (unsigned char *)ptr;
    register unsigned char *end = p + s;
    register int ch;

    for ( ; p < end; p++, i++ ) {
        if ( (i % 32) == 0 ) {
            fprintf(fp,"    ");
        }
        ch = *p;
        if ( isprint(ch) ) {
            fputc('.',fp);
            fputc(ch,fp);
        } else {
            fprintf(fp,"%02x",ch);
        }
        if ( ((i+1) % 32) == 0 ) {
            fputc('\n',fp);
        }
    }
    fputc('\n',fp);
}

static void dmem_report(FILE *fp)
{
    udict_node_t *n;
    DMEM_T *d;
    size_t leaks = 0;

    if ( initialized != INITIALIZED ) return;

    fprintf(fp,"memory leaks: %ld memory chunks have not been freed\n",udict_count(dict));

    n = udict_first(dict);
    while ( n != NULL ) {
        d = (DMEM_T *)udict_node_getdata(n);

        if ( dumpLeaks )
            fprintf(fp,"0x%08x(%d bytes): allocated at %s(%d)\n",(unsigned int)d->addr,d->size,d->file,d->line);

        leaks += d->size;

        if ( dumpMemory )
            dump_memory(fp,d->addr,d->size);

        n = udict_next(dict,n);
    }

    fprintf(fp,"Total: %d bytes memory leaks\n",leaks);
}

void *dmem_malloc(const char *fn,int l,size_t s)
{
    void *p = NULL;

    if ( initialized == NOT_INITIALIZED ) {
        dmem_initialize();
    }

    p = malloc(s);

    if ( initialized == INITIALIZED && p != NULL ) {
        /* put DMEM_T into the list */
        DMEM_T *d = dmem_t_new(fn,l,p,s);
        if ( d == NULL || !udict_alloc_insert(dict,d->addr,d) ) {
            free(p);
            p = NULL;
            if ( d != NULL ) dmem_t_free(d);
        }
    }
    return p;
}

void *dmem_calloc(const char *fn,int l,size_t n,size_t s)
{
    void *p = NULL;

    if ( initialized == NOT_INITIALIZED ) {
        dmem_initialize();
    }

    p = calloc(n,s);

    if ( initialized == INITIALIZED && p != NULL ) {
        /* put DMEM_T into the list */
        DMEM_T *d = dmem_t_new(fn,l,p,n*s);
        if ( d == NULL || !udict_alloc_insert(dict,d->addr,d) ) {
            free(p);
            p = NULL;
            if ( d != NULL ) dmem_t_free(d);
        }
    }
    return p;
}


void *dmem_realloc(const char *fn,int l,void *ptr,size_t s)
{
    udict_node_t *n = NULL;
    DMEM_T *m = NULL;

    void *p = NULL;
    if ( initialized == NOT_INITIALIZED ) {
        dmem_initialize();
    }
    if ( initialized != INITIALIZED ) {
        return realloc(ptr,s);
    }

    if ( ptr != NULL ) {
        n = udict_lookup(dict,ptr);
        if ( n == NULL ) {
            fprintf(reportFp,"%s(%d): invalid realloc() against memory 0x%08x(%d bytes)\n",fn,l,(unsigned int)ptr,s);
            exit(1);
        }
        m = udict_node_getdata(n);
    } else if ( s == 0 ) {
        fprintf(reportFp,"%s(%d): invalid realloc(). Should be given valid memory\n",fn,l);
        exit(1);
    }

    p = realloc(ptr,s);

    if ( initialized != INITIALIZED )
        return p;

    if ( n == NULL ) { /* if ptr == NULL, same as malloc */
        m = dmem_t_new(fn,l,p,s);
        if ( m == NULL || !udict_alloc_insert(dict,m->addr,m) ) {
            free(p);
            p = NULL;
            if ( m != NULL ) dmem_t_free(m);
        }
    } else if ( s == 0 ) { /* if s == 0, same as free */
        m = udict_node_getdata(n);
        udict_delete_free(dict,n);
        dmem_t_free(m);
    } else { /* if p == ptr, only size has been changed */
        if ( p == ptr ) {
            m->size = s;
        } else {
            udict_delete(dict,n);
            m->addr = p;
            m->file = fn;
            m->line = l;
            m->size = s;
            udict_insert(dict,n,p);
        }
    }

    return p;
}

char *dmem_strdup(const char *fn,int l,const char *s)
{
    char *p = NULL;
    if ( initialized == NOT_INITIALIZED ) {
        dmem_initialize();
    }

    p = strdup(s);

    if ( initialized == INITIALIZED && p != NULL ) {
        /* put DMEM_T into the list */
        DMEM_T *d = dmem_t_new(fn,l,p,strlen(p)+1);
        if ( d == NULL || !udict_alloc_insert(dict,d->addr,d) ) {
            free(p);
            p = NULL;
            if ( d != NULL ) dmem_t_free(d);
        }
    }
    return p;
}

void dmem_free(const char *fn,int l,void *ptr)
{
    if ( ptr == NULL ) return; /* by default */

    if ( initialized == INITIALIZED ) {
        /* check whether ptr is valid */
        udict_node_t *n = udict_lookup(dict,ptr);
        if ( n == NULL ) {
            /* invalid free. insert log?  */
            fprintf(reportFp,"%s(%d): invalid free() against memory 0x%08x\n",fn,l,(unsigned int)ptr);
            exit(1);
        } else {
            DMEM_T *t = (DMEM_T *)udict_node_getdata(n);
            udict_delete_free(dict,n); 
            dmem_t_free(t); /* destroy data object */
        }
    }
    free(ptr);
}

